/*
	This MUST be compiled with Borland Turbo C++ 3.0 in a Tiny memory model
	because it needs CS == DS and makes assumptions about the function layout.

	I might have written it entirely in assembly language, but this way I
	don't have to worry about command line parsing or string output.

    Since the program doesn't have its own stack running, it borrows the
    stack of the application calling it, using only a few bytes (less than
    32). This may cause corruption in programs running under *very* tight
    conditions, but should not realistically be a problem in all but the
    most exotic cases.
*/
#ifndef __TURBOC__
#error This MUST be compiled with Borland Turbo C++ 3.0.
#endif
#ifndef __TINY__
#error This MUST be compiled with Tiny memory model.
#endif

#include <dos.h>

//Need little memory (128 bytes will do)
extern unsigned _heaplen = 64;
extern unsigned _stklen = 64;

const char far *src = "";
const char far *tgt = "";

void far *oldDOS;
void interrupt (*newDOS)(__CPPARGS);

void interrupt_handler (void)
{
	asm {
		//Set up stack frame
		push bp
		mov bp, sp

		//Is it "OPEN EXISTING FILE"?
        pushf
		cmp ah, 0x3D
		je process
        popf

		//No, just call old handler
        pushf
		call dword ptr [cs:oldDOS]

		//Update flags on stack
		sti
		push dx
		pushf
		pop dx
		mov word ptr [ss:bp+0x06], dx
		pop dx
		pop bp

		//And be done with it
		iret
	}
process:
	asm {
    	popf

		//Store arguments
		push ds
		push dx
		push es
		push si
		push di
		push ax

		//Get pointers
		push ds
		mov ax, cs
		mov ds, ax
		les di, dword ptr src
		pop ds
		mov si, dx
	}
cmpchar:
	asm {
		//Read one character from both strings
		mov ah, byte ptr [ds:si]
		mov al, byte ptr [es:di]

		//Convert to uppercase
		cmp ah, 'a'
		jl a_ok
		cmp ah, 'z'
		jg a_ok
		and ah, 0xDF
	}
a_ok:
	asm {
		cmp al, 'a'
		jl b_ok
		cmp al, 'z'
		jg b_ok
		and al, 0xDF
	}
b_ok:
	asm {
		//Different? Then call the old handler unchanged
		cmp al, ah
		jne callold

		inc si
		inc di

		//Non-zero? Then keep comparing
		test al, al
		jnz cmpchar

		//Zero, use substituted filename
		mov ax, cs
		mov ds, ax
		lds dx, dword ptr tgt
	}
callold:
	asm {
		//Call old handler
		pop ax
		pop di
		pop si
		pop es
		pushf
		call dword ptr [cs:oldDOS]

		//Update flags on stack
		sti
		pushf
		pop dx
		mov word ptr [ss:bp+0x06], dx

		//Restore values
		pop dx
		pop ds

		//All done
		pop bp
		iret
	}
}

int main (int argc, const char **argv)
{
	//Must be two arguments (plus program name makes 3)
	if (argc != 3)
	{
		printf ("Usage: fsubst <src> <tgt>\n");
		return 1;
	}

	//Filename to replace
	src = argv [1];

	//Substitute filename
	tgt = argv [2];

	//Ugly fudge of handler address (assumes inline ASM starts after 5 bytes)
	((void far *)newDOS) = interrupt_handler;
	((char far *)newDOS) += 5;

	//Switch interrupt vectors
	oldDOS = getvect (0x21);
	setvect (0x21, newDOS);

	//Stay resident (an extra 256 bytes to ensure parameters are still loaded)
	keep (0, (_SS + ((_SP + 256) >> 4) - _psp));

	return 0;
}
